use std::sync::Arc;

use sqlx::PgPool;

use domain::models::todo::{CreateTodo, Todo};
use domain::repositories::repository::{RepositoryResult, ResultPaging, RepositoryOperation, RepositoryObject};
use domain::repositories::todo::TodoRepository;
use domain::models::Id;
use crate::repositories::repository::PostgresRepositoryError;

#[derive(Clone)]
pub struct TodoPostgresRepository {
    pub pool: Arc<PgPool>
}


impl TodoPostgresRepository {
    pub fn new(pool: Arc<PgPool>) -> Self {
        TodoPostgresRepository { pool }
    }
}

impl TodoRepository for TodoPostgresRepository {

    async fn create(&self, new_todo: &CreateTodo) -> RepositoryResult<Todo> {
        let title = &new_todo.title();
        let description = &new_todo.description();
        let record = sqlx::query_as!(Todo,
            r#"INSERT INTO todos
            (title, description)
            VALUES ($1, $2)
            RETURNING id, title, description, completed"#,
            title,
            description
        ).fetch_one(&*self.pool)
            .await
            .map_err(|v| PostgresRepositoryError::from((v, RepositoryOperation::Create, RepositoryObject::Todo)).into_inner())?;

        Ok(record)
    }

    async fn list(&self) -> RepositoryResult<ResultPaging<Todo>> {
        let records = sqlx::query_as!(Todo,
            "SELECT * FROM todos ORDER BY id"
        ).fetch_all(&*self.pool)
            .await
            .map_err(|v| PostgresRepositoryError::from((v, RepositoryOperation::GetMany, RepositoryObject::Todo)).into_inner())?;

        Ok(ResultPaging {
            total: records.len().try_into().unwrap(),
            items: records.into_iter().map(|v| v.into()).collect()
        })
    }

    async fn get(&self, id: Id) -> RepositoryResult<Todo> {
        let id = id.id();
        let record = sqlx::query_as!(Todo,
            "SELECT id, title, description, completed FROM todos WHERE id = $1",
            id
        ).fetch_one(&*self.pool)
            .await
            .map_err(|v| PostgresRepositoryError::from((v, RepositoryOperation::GetOne, RepositoryObject::Todo)).into_inner())?;

        Ok(record)
    }

    async fn delete(&self, id: Id) -> RepositoryResult<()> {
        let id = id.id();
        let _ = sqlx::query_as!(Todo,
            "DELETE FROM todos WHERE id = $1",
            id
        ).execute(&*self.pool)
            .await
            .map_err(|v| PostgresRepositoryError::from((v, RepositoryOperation::Delete, RepositoryObject::Todo)).into_inner())?;
        
        Ok(())
    }

    async fn update(&self, todo: &Todo) -> RepositoryResult<Todo> {
        let id = todo.id().id();
        let title = todo.title();
        let description = todo.description();
        let completed = todo.completed();
        let record = sqlx::query_as!(
            Todo,
            r#"--sql
            with update_todo as (
                update todos
                set title = $2, description = $3, completed = $4
                where id = $1
                returning id, title, description, completed
            )
            select * from update_todo
            "#,
            id,
            title,
            description,
            completed
        ).fetch_one(&*self.pool)
            .await
            .map_err(|v| PostgresRepositoryError::from((v, RepositoryOperation::Update, RepositoryObject::Todo)).into_inner())?;
        
        Ok(record)
    }

}
