//! A generalised ID type for entities and aggregates.

use std::fmt;
use std::fmt::Display;
use serde::Serialize;

#[derive(Debug, Serialize)]
pub struct Id(uuid::Uuid);

#[derive(Debug, Clone)]
pub struct IdParseError {}

impl Id {
    pub fn new() -> Self {
        Self {
            0: uuid::Uuid::now_v7(),
        }
    }

    pub fn from_string(from: String) -> Result<Self, IdParseError> {
        uuid::Uuid::parse_str(from.as_str())
            .map_err(|_| IdParseError {})
            .map(|val| Self {
                0: val
            })
    }

    pub fn to_string(self) -> String {
        self.0.to_string()
    }

    pub fn id(self) -> uuid::Uuid {
        self.0
    }
}

impl From<uuid::Uuid> for Id {
    fn from(uuid: uuid::Uuid) -> Self {
        Self { 0: uuid }
    }
}

impl Clone for Id {
    fn clone(&self) -> Self {
        *self
    }
}

impl Copy for Id {}

impl PartialEq for Id {
    fn eq(&self, other: &Self) -> bool {
        self.0 == other.0
    }
}

impl Display for Id {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.0)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn format_id() {
        let uuid_value = uuid::Uuid::now_v7();
        let id = Id::from_string(uuid_value.to_string()).expect("");
        assert_eq!(format!("{id}"), uuid_value.to_string());
    }
}
